package org.jboss.wiki;

import java.io.IOException;
import java.util.HashMap;
import java.util.InvalidPropertiesFormatException;
import java.util.Iterator;
import java.util.Properties;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.jboss.logging.Logger;
import org.jboss.wiki.plugins.HTMLTranslatorParts;

public class InsidePluginExcecutor {

	private static final String attribType1 = "[\\s]+[\\w]+=([\\d]+|true|false),?";

	private static final String attribType2 = "[\\s]+[\\w]+='(\\\\'|[^'])+',?";

	private static final String insidePluginRegex = "\\{(INSERT[\\s]){0,1}[\\s]*[.\\w]+([\\s]+WHERE[\\s]*)?("
			+ attribType1 + "|" + attribType2 + ")*[\\s]*\\}";

	private HashMap<String, WikiInsidePlugin> insidePlugins;

	private Logger log;

	private WikiEngine wikiEngine;

	public InsidePluginExcecutor(WikiEngine wikiEngine) {

		this.wikiEngine = wikiEngine;

		log = Logger.getLogger(this.getClass());

		insidePlugins = new HashMap<String, WikiInsidePlugin>();

		loadInsidePlugins();
	}

	private void loadInsidePlugins() {
		Properties props = new Properties();

		try {
			props.loadFromXML(WikiTypeLoader.class
					.getResourceAsStream("/WikiInsidePlugins.xml"));
		} catch (InvalidPropertiesFormatException e) {
			log.error("WikiInsidePlugins.xml has bad format !", e);
		} catch (IOException e) {
			log.error("There is no WikiInsidePlugins.xml file !", e);
		}

		Iterator it = props.keySet().iterator();

		while (it.hasNext()) {
			String pluginName = (String) it.next();
			log.info("Loading inside plugin: " + pluginName + " class: "
					+ props.getProperty(pluginName));

			WikiInsidePlugin wikiInsidePlugin = null;

			try {
				wikiInsidePlugin = loadObject(props.getProperty(pluginName));

				insidePlugins.put(pluginName, wikiInsidePlugin);

			} catch (InstantiationException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (ClassNotFoundException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			}
		}

	}

	private WikiInsidePlugin loadObject(String className)
			throws InstantiationException, IllegalAccessException,
			ClassNotFoundException {
		WikiInsidePlugin wikiInsidePlugin =  (WikiInsidePlugin) Class.forName(className).newInstance();
		
		wikiInsidePlugin.setWikiEngine(wikiEngine);
		
		return wikiInsidePlugin;
	}

	public String executeInsidePlugin(String pluginLink, WikiPage wikiPage,
			WikiSession wikiSession) {

		// old inside plugin regex
		// "\\{[\\w]+([\\s]+[\\w]+='[\\d\\w\\s]+')*\\s*\\}"

		String localPluginLink = pluginLink;
		
		WikiInsidePlugin plugin = null;

		log.debug("executing plugin: "+pluginLink);
		
		if (!localPluginLink.matches(insidePluginRegex)) {
			// this is not a proper inside plugin link

			return "''INSIDE PLUGIN ERROR: " + pluginLink
					+ " is not a proper plugin link''";
		}

		plugin = getPluginObject(localPluginLink);

		if (plugin != null) {
			HashMap<String, String> attributes = extractAttribs(localPluginLink);
			
			return (plugin.processContent() ? HTMLTranslatorParts.pagePartRegexStart : "") + plugin.process(wikiPage, wikiSession,
					attributes) + (plugin.processContent() ? HTMLTranslatorParts.pagePartRegexEnd : "");
		}
		else
			return "''INSIDE PLUGIN ERROR: There is no plugin: " + pluginLink
					+ "''";

	}

	private WikiInsidePlugin getPluginObject(String pluginDef) {
		WikiInsidePlugin wikiInsidePlugin = null;

		if (pluginDef.startsWith("{"))
			pluginDef = pluginDef.substring(1);
		
		pluginDef = pluginDef.trim();

		if (pluginDef.startsWith("INSERT")) {
			pluginDef = pluginDef.substring("INSERT".length());
		}

		pluginDef = pluginDef.trim();

		int spacePos = pluginDef.indexOf(" ");
		
		String className = pluginDef.substring(0, (spacePos != -1) ? spacePos : pluginDef.indexOf("}"));

		pluginDef = pluginDef.substring(className.length()).trim();

		log.debug("Plugin name: "+className);
		
		if (className.indexOf('.') != -1) {
			// full class name
			try {
				wikiInsidePlugin = loadObject(className);
			} catch (InstantiationException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				// TODO Auto-generated catch block
				e.printStackTrace();
			} catch (ClassNotFoundException e) {
				// just return null
			}
		} else {
			wikiInsidePlugin = insidePlugins.get(className);
		}

		return wikiInsidePlugin;
	}

	private HashMap<String, String> extractAttribs(String pluginDef) {
		
		HashMap<String, String> attributes = new HashMap<String, String>();

		Matcher matcher = Pattern.compile(attribType1).matcher(pluginDef);

		while (matcher.find()) {
			String attrib = pluginDef.substring(matcher.start(), matcher.end());

			getAttrib1(attrib.trim(), attributes);
		}
		
		matcher = Pattern.compile(attribType2).matcher(pluginDef);

		while (matcher.find()) {
			String attrib = pluginDef.substring(matcher.start(), matcher.end());

			getAttrib2(attrib.trim(), attributes);
		}

		return attributes;
	}
	
	private void getAttrib1(String attrib, HashMap<String,String> attributes) {
		// digits or true/false
		String attribName = attrib.substring(0, attrib.indexOf("="));
		String attribValue = attrib.substring(attrib.indexOf("=")+1, attrib.length());
		
		attributes.put(attribName, attribValue);
		
		log.debug("new attribute: "+attribName+" "+attribValue);
	}
	
	private void getAttrib2(String attrib, HashMap<String,String> attributes) {
		// string, digits ... also special characters (\' = ')
		
		String attribName = attrib.substring(0, attrib.indexOf("="));
		String attribValue = attrib.substring(attrib.indexOf('\'')+1, attrib.length()-1).replaceAll("\\\\'", "'");
		
		attributes.put(attribName, attribValue);
		
		log.debug("new attribute: "+attribName+" "+attribValue);
	}
}
