/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jboss.wiki;

import java.io.IOException;
import java.io.InputStream;
import java.util.Arrays;
import java.util.Comparator;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Map.Entry;
import java.util.concurrent.Semaphore;

import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import org.jboss.wiki.exceptions.DataSourceException;
import org.jboss.wiki.exceptions.EditingNotAllowedException;
import org.jboss.wiki.exceptions.PageNotEditedException;
import org.jboss.wiki.exceptions.PageRenamingException;
import org.jboss.wiki.exceptions.WikiSaveException;
import org.jboss.wiki.plugins.WatchListPlugin;

import org.jboss.logging.Logger;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.xml.sax.SAXException;

/**
 * 
 * @author <a href="mailto:dsicore@jboss.org">Damon Sicore</a>
 * @author <a href="mailto:tomasz.szymanski@jboss.com">Tomasz Szymanski</a>
 * @version $Id$
 */
public class WikiEngine {

	private static final String LANGUAGE_CODES_CLASS = "CountryCodes.xml";

	/**
	 * <p>
	 * </p>
	 * 
	 */
	public static String defaultPage = "Main";

	public static final String PROPERTIES_FILENAME = "JBossWiki.properties";

	public static final String DEFAULT_LANGUAGE_PROP = "defaultLanguageCode";

	public String defaultLangugeCode = "EN";

	private HashMap<String, WikiType> wikiTypes;

	private WikiPage noSuchPage;

	private Integer sessionIdExpireTime = 120;

	private int editSessionId = 0;

	private Set<Integer> editedPages;

	// private Semaphore s = new Semaphore(1);

	private Class credentialsClass;

	private Logger log;

	private InsidePluginExcecutor insidePluginExcecutor;

	private Map<String, String> allLanguageCodes;

	private Map<String, String> usedLanguageCodes;

	private Map<String, LanguageDataSource> languageDataSources;

	public boolean pageExists(String pageName, String langCode) {
		return languageDataSources.get(langCode).getMediaDataSource()
				.pageExists(getUid(pageName, langCode));
	}

	/**
	 * <p>
	 * Gets page with default languiage code.
	 * </p>
	 * 
	 * 
	 * @param pageName
	 * @return
	 * @param wikiContext
	 */
	public synchronized WikiPage getByName(String pageName,
			WikiContext wikiContext, String langCode) {
		return getByName(pageName, wikiContext, languageDataSources
				.get(langCode));
	}

	private LanguageDataSource defaultDS() {
		return languageDataSources.get(defaultLangugeCode);
	}

	public synchronized WikiPage getByName(String pageName,
			WikiContext wikiContext, LanguageDataSource dataSource) {

		/*
		 * try { s.acquire(); } catch (InterruptedException e) {
		 * e.printStackTrace(); }
		 */

		String realPageName = getUid(pageName, dataSource.getLanguageCode());

		WikiPage ret = null;

		if (!dataSource.getMediaDataSource().preGet()) {
			log.error("Couldn't get page");
			// s.release();
			return null;
		}

		if (dataSource.getPages().containsKey(realPageName)) {
			ret = dataSource.getPages().get(realPageName);
		} else {
			ret = dataSource.getMediaDataSource().getPage(realPageName);

			if (ret != null) {
				ret.setName(pageName);

				ret.setLangCode(dataSource.getMediaDataSource().getLangCode());

				dataSource.getPages().put(realPageName, ret);
			}
		}

		if (((wikiContext != null) && (ret != null)
				&& (!ret.isViewable(wikiContext.getUser())) && (!wikiContext
				.getUser().isAdmin()))) {
			// s.release();
			ret = new WikiPage(realPageName, null,
					"I am sorry, but you are not allowed to see this page", 0,
					0, new Date(), this, false, false, null, null, 0,
					dataSource.getMediaDataSource().getLangCode());

			refreshPage(pageName, dataSource.getLanguageCode());

			return ret;
		}

		if (!dataSource.getMediaDataSource().postGet()) {
			log.error("Problems with postGet");
		}

		if (ret != null) {
			if (wikiContext != null) {
				// s.release();
				return wikiContext.process(ret);
			} else {
				// s.release();
				return ret;
			}
		} else {
			try {
				ret = (WikiPage) noSuchPage.clone();
				ret.setName(pageName);
				ret.setLangCode(dataSource.getMediaDataSource().getLangCode());
			} catch (CloneNotSupportedException cne) {
				log.error(cne);

			}
			if (wikiContext != null) {
				// s.release();
				return wikiContext.process(ret);
			} else {
				// s.release();
				return ret;
			}
		}
	}

	public synchronized WikiPage getByName(String pageName,
			WikiContext wikiContext, int version, String langCode) {
		return getByName(pageName, wikiContext, version, languageDataSources
				.get(langCode));
	}

	public synchronized WikiPage getByName(String pageName,
			WikiContext wikiContext, int version, LanguageDataSource dataSource) {
		/*
		 * try { s.acquire(); } catch (InterruptedException e) {
		 * e.printStackTrace(); }
		 */

		String realPageName = getUid(pageName, dataSource.getLanguageCode());

		WikiPage ret = null;

		if (dataSource.getPages().containsKey(realPageName)) {
			ret = dataSource.getPages().get(realPageName);
		} else {
			ret = dataSource.getMediaDataSource().getPage(realPageName);

			ret.setLangCode(dataSource.getMediaDataSource().getLangCode());

			ret.setName(pageName);

			dataSource.getPages().put(realPageName, ret);
		}

		ret = ret.getPageAtVersion(true, version);

		if (ret != null) {
			if (wikiContext != null) {
				// s.release();
				return wikiContext.process(ret);
			} else {
				// s.release();
				return ret;
			}
		} else {
			// return noSuchPage if there isn't any
			/*
			 * try { ret = (WikiPage) noSuchPage.clone(); } catch
			 * (CloneNotSupportedException cne) { log.error(cne); } if
			 * (wikiContext != null) { return wikiContext.process(ret); } else {
			 * return ret; }
			 */
			// s.release();
			return null;
		}
	}

	private void loadLanguageCodes() {
		allLanguageCodes = new TreeMap<String, String>();

		usedLanguageCodes = new TreeMap<String, String>();

		languageDataSources = new TreeMap<String, LanguageDataSource>();

		try {
			Document languageXML = DocumentBuilderFactory.newInstance()
					.newDocumentBuilder().parse(
							WikiEngine.class.getResourceAsStream("/"
									+ LANGUAGE_CODES_CLASS));

			NodeList nodes = languageXML.getElementsByTagName("country");

			for (int i = 0; i < nodes.getLength(); i++) {
				Node node = nodes.item(i);

				NodeList countryNodes = node.getChildNodes();

				String desc = null, twoLetterCode = null;

				for (int j = 0; j < countryNodes.getLength(); j++) {
					Node countryNode = countryNodes.item(j);

					if (countryNode.getNodeName().equals("description")) {
						desc = unmarshallText(countryNode);
					} else if (countryNode.getNodeName()
							.equals("twoLetterCode")) {
						twoLetterCode = unmarshallText(countryNode);
					}
				}

				if (desc != null || twoLetterCode != null)
					allLanguageCodes.put(twoLetterCode, desc);
				else {
					log.error("Problems with country code. Code: "
							+ twoLetterCode + " country: " + desc);
				}
			}

		} catch (SAXException e) {
			log.error("Couldn't parse language codes file", e);
		} catch (IOException e) {
			log.error("Couldn't parse language codes file", e);
		} catch (ParserConfigurationException e) {
			log.error("Couldn't parse language codes file", e);
		}

		log.debug("Country codes read from CountryCodes.xml:");
		for (String country : allLanguageCodes.keySet()) {
			log.debug("Country: " + country + " : "
					+ allLanguageCodes.get(country));
		}
	}

	/**
	 * <p>
	 * Does ...
	 * </p>
	 * 
	 * 
	 * @return
	 */
	public void init() {
		log = Logger.getLogger(WikiEngine.class);

		editedPages = new HashSet<Integer>();

		wikiTypes = new HashMap<String, WikiType>();

		noSuchPage = new WikiPage("", new SimpleCredentials(""),
				"There is no such page. Click on EDIT to start it.", 0, 0,
				null, this, true, true, null, null, 0, defaultLangugeCode);

		loadWikiTypes();
		loadLanguageCodes();

		insidePluginExcecutor = new InsidePluginExcecutor(this);

		Properties jbosswikiprops = new Properties();

		try {
			jbosswikiprops.load(WikiEngine.class.getResourceAsStream("/"
					+ PROPERTIES_FILENAME));
		} catch (IOException e) {
			e.printStackTrace();
		}

		// load all properties
		if (jbosswikiprops.getProperty("defaultPage") != null) {
			defaultPage = jbosswikiprops.getProperty("defaultPage");
		}

		if (jbosswikiprops.getProperty("sessionIdExpireTime") != null) {
			sessionIdExpireTime = Integer.valueOf(jbosswikiprops
					.getProperty("sessionIdExpireTime"));
		}

		if (jbosswikiprops.getProperty(DEFAULT_LANGUAGE_PROP) != null) {
			defaultLangugeCode = jbosswikiprops
					.getProperty(DEFAULT_LANGUAGE_PROP);
		}

		// add default language
		usedLanguageCodes.put(defaultLangugeCode, allLanguageCodes
				.get(defaultLangugeCode));

		/*
		 * mediaDataSource = new FileDataSource(); attachementDataSource = new
		 * FileDataSource(); wikiPageDictionary = new FileDataSource();
		 * 
		 * 
		 */

		MediaDataSource mediaDataSource = null;

		AttachmentDataSource attachmentDataSource = null;

		WikiPageDictionary wikiPageDictionary = null;

		// LOAD data sources
		if (jbosswikiprops.getProperty("mediaDataSourceClass") != null) {
			try {
				mediaDataSource = (MediaDataSource) Class.forName(
						jbosswikiprops.getProperty("mediaDataSourceClass"))
						.newInstance();
			} catch (InstantiationException e) {
				log.error(e);
			} catch (IllegalAccessException e) {
				log.error(e);
			} catch (ClassNotFoundException e) {
				log.error(e);
			}
			log.info("Loaded mediaDataSource: " + mediaDataSource);

		} else {
			// if sth went wrong
			mediaDataSource = new FileMediaDataSource();
		}

		if (jbosswikiprops.getProperty("attachmentDataSourceClass") != null) {
			try {
				attachmentDataSource = (AttachmentDataSource) Class
						.forName(
								jbosswikiprops
										.getProperty("attachmentDataSourceClass"))
						.newInstance();
			} catch (InstantiationException e) {
				log.error(e);
			} catch (IllegalAccessException e) {
				log.error(e);
			} catch (ClassNotFoundException e) {
				log.error(e);
			}

			log.info("Loaded attachmentDataSource: " + attachmentDataSource);
		} else {
			// if sth went wrong
			attachmentDataSource = new FileAttachmentDataSource();
		}

		if (jbosswikiprops.getProperty("wikiPageDictionaryClass") != null) {
			try {
				wikiPageDictionary = (WikiPageDictionary) Class.forName(
						jbosswikiprops.getProperty("wikiPageDictionaryClass"))
						.newInstance();
			} catch (InstantiationException e) {
				log.error(e);
			} catch (IllegalAccessException e) {
				log.error(e);
			} catch (ClassNotFoundException e) {
				log.error(e);
			}
			log.info("Loaded wikiPageDictionary: " + wikiPageDictionary);
		} else {
			// if sth went wrong
			wikiPageDictionary = new FileWikiPageDictionary();
		}

		// set wikiEngine in classes
		mediaDataSource.setWikiEngine(this);
		attachmentDataSource.setWikiEngine(this);
		wikiPageDictionary.setWikiEngine(this);

		// create language data source pack for default language
		LanguageDataSource languageDataSource = new LanguageDataSource(this,
				defaultLangugeCode, mediaDataSource, attachmentDataSource,
				wikiPageDictionary);
		languageDataSources.put(defaultLangugeCode, languageDataSource);

		// load credential class
		if (jbosswikiprops.getProperty("credentialsClass") != null) {
			try {
				credentialsClass = Class.forName(jbosswikiprops
						.getProperty("credentialsClass"));
			} catch (ClassNotFoundException e) {
				log.error("Coudln't load credentials class.", e);
			}

			if (credentialsClass != null)
				log.info("Loaded credentialsClass: " + credentialsClass);
		}

		// if credentialClass didn't load or wasn't set
		if (credentialsClass == null) {
			credentialsClass = SimpleCredentials.class;
			log.info("(SIMPLE)Loaded credentialsClass: " + credentialsClass);
		}

		// load all languageDSrces (except the default loaded one)

		Map<String, String> langs = defaultDS().getMediaDataSource()
				.getAvailableLanguages();

		for (String lang : langs.keySet()) {
			if (!lang.equals(defaultLangugeCode))
				loadLanguage(lang);
		}

	}

	/**
	 * <p>
	 * Parses configuration file to get WikiTypes
	 * </p>
	 * 
	 * 
	 * @return
	 */
	private void loadWikiTypes() {
		try {
			wikiTypes.clear();

			WikiTypeLoader wikiTypeLoader = new WikiTypeLoader(this);

			wikiTypeLoader.parse(WikiTypeLoader.class
					.getResourceAsStream("/WikiTypes.xml"));

		} catch (Exception se) {
			log.error("Error while parsing WikiType configuration: ", se);
		}
	}

	public void addWikiType(String name, WikiType wikiType) {
		log.info("Added WikiType: " + name + " Class: " + wikiType);
		wikiTypes.put(name, wikiType);
	}

	public WikiType getWikiType(String wikiTypeName) {
		// System.out
		// .println("[WIKITYPE REQUEST]: " + wikiTypes.get(wikiTypeName));
		return wikiTypes.get(wikiTypeName);
	}

	/**
	 * <p>
	 * </p>
	 * 
	 */
	public WikiEngine() {
		init();
	}

	public void refreshPage(String pageName, String langCode) {
		if (languageDataSources.get(langCode).getPages().containsKey(pageName)) {
			languageDataSources.get(langCode).getPages().remove(pageName);
		}
	}

	public Map<String, Integer> findPages(String searchQuery, String langCode) {
		// log.info("Starting search: " + searchQuery);
		WikiSearch wikiSearch = new WikiSearch(searchQuery);

		Map<String, Integer> resPages = new HashMap<String, Integer>();

		log.debug(langCode);

		Set<String> pages = languageDataSources.get(langCode)
				.getMediaDataSource().getAllPageNames();

		for (String name : pages) {
			WikiPage temp = getByName(name, null, langCode);

			if (temp != null) {
				int searchScore;

				// search for pages (add pagename at the beginig of each page so
				// it can be found)
				if ((searchScore = wikiSearch.matches((name + " " + temp
						.getPageContent()).toLowerCase())) > 0) {
					resPages.put(name, searchScore);
				}
			}
		}

		Object[] entryArr = resPages.entrySet().toArray();
		// Sorting the resPages by entry values (which are numbers), descending.
		Arrays.sort(entryArr, new Comparator<Object>() {
			public int compare(Object entry1, Object entry2) {
				int i1 = ((Entry<String, Integer>) entry1).getValue();
				int i2 = ((Entry<String, Integer>) entry2).getValue();

				if (i1 == i2)
					return 0;
				return (i1 < i2) ? 1 : -1;
			}
		});

		resPages = new LinkedHashMap<String, Integer>();

		for (int i = 0; i < entryArr.length; i++) {
			Entry<String, Integer> entry = (Entry<String, Integer>) entryArr[i];

			resPages.put(entry.getKey(), entry.getValue());
		}
		// ("Search completed");
		return resPages;
	}

	public Set<String> getPagesFor(String pageName, String langCode) {
		Set<String> set = languageDataSources.get(langCode)
				.getMediaDataSource().getPagesFor(pageName);
		Set<String> ret = new HashSet<String>();

		for (String page : set) {
			ret.add(getRealName(page, langCode));
		}

		return ret;
	}

	public String executeInsidePlugin(String pluginLink, WikiPage wikiPage,
			WikiSession wikiSession) {

		return insidePluginExcecutor.executeInsidePlugin(pluginLink, wikiPage,
				wikiSession);

	}

	public boolean deletePage(String pageName, boolean deleteSubpages,
			String langCode) {
		Set<String> subpages = new HashSet<String>();

		MediaDataSource mediaDataSource = languageDataSources.get(langCode)
				.getMediaDataSource();
		AttachmentDataSource attachmentDataSource = languageDataSources.get(
				langCode).getAttachmentDataSource();

		log.info("DELETE SUBPAGES: " + deleteSubpages);

		if (deleteSubpages) {
			subpages = mediaDataSource.getPagesFor(getUid(pageName, langCode));
		}

		if (attachmentDataSource.getAttachmentsSet(getByName(
				getUid(pageName, langCode), null, langCode).getName()) != null) {
			if (!attachmentDataSource.deleteAttachments(pageName)) {
				log.error("Couldn't delete attachments of page " + pageName);
				return false;
			}
		}

		boolean ret = true;

		if ((deleteSubpages) && (subpages.size() > 0)) {

			for (String subPage : subpages) {
				if (!deletePage(subPage, deleteSubpages, langCode)) {
					log.error("[WikiEngine]: Problems with deleteing subpage: "
							+ subPage);
					ret = false;
				}
			}

		}

		if (deletePage(pageName, langCode)) {

			if (languageDataSources.get(langCode).getPages().containsKey(
					pageName)) {
				languageDataSources.get(langCode).getPages().remove(pageName);
			}

			log.info("Finishing deleting of page " + pageName + " with ret = "
					+ ret);

			return ret;
		} else {
			log.error("Problems with deleting page " + pageName);
			return false;
		}
	}

	public String getRealName(String uid, String langCode) {
		String[] pageTokens = uid.split("/");

		// log.info("Looking at page: (realName) "+uid);

		if (pageTokens.length > 1) {
			if (!languageDataSources.get(langCode).getWikiPageDictionary()
					.getRealName(uid).equals(uid)) {

				// this page is in dictionary
				return languageDataSources.get(langCode)
						.getWikiPageDictionary().getRealName(uid);
			} else {
				// log.info("checking parts");
				// look if parts of this page are in dictionary

				for (int i = pageTokens.length - 1; i > 0; i--) {
					String page = "";
					String suffix = "";
					for (int j = 0; j < i; j++) {
						page += "/" + pageTokens[j];
					}

					// build suffix

					for (int j = i; j < pageTokens.length; j++) {
						suffix += "/" + pageTokens[j];
					}

					page = page.substring(1);

					// log.info("Looking at page: (realName) "+page);

					if (!languageDataSources.get(langCode)
							.getWikiPageDictionary().getRealName(page).equals(
									page)) // this
						// page
						// is
						// in
						// dictionary
						return languageDataSources.get(langCode)
								.getWikiPageDictionary().getRealName(page)
								+ suffix;

				}

				return uid;
			}
		} else
			return languageDataSources.get(langCode).getWikiPageDictionary()
					.getRealName(uid);
	}

	public String getUid(String realName, String langCode) {
		String[] pageTokens = realName.split("/");

		// log.info("Looking at page: (uid) "+realName);

		if (pageTokens.length > 1) {
			if (!languageDataSources.get(langCode).getWikiPageDictionary()
					.getUid(realName).equals(realName)) {
				// this page is in dictionary
				return languageDataSources.get(langCode)
						.getWikiPageDictionary().getUid(realName);
			} else {
				// look if parts of this page are in dictionary

				for (int i = pageTokens.length - 1; i > 0; i--) {
					String page = "";
					String suffix = "";
					for (int j = 0; j < i; j++) {
						page += "/" + pageTokens[j];
					}

					// build suffix

					for (int j = i; j < pageTokens.length; j++) {
						suffix += "/" + pageTokens[j];
					}

					page = page.substring(1);

					// log.info("Looking at page: (uid) "+page);

					if (!languageDataSources.get(langCode)
							.getWikiPageDictionary().getUid(page).equals(page)) // this
						// page
						// is
						// in
						// dictionary
						return languageDataSources.get(langCode)
								.getWikiPageDictionary().getUid(page)
								+ suffix;
				}

				return realName;
			}
		} else
			return languageDataSources.get(langCode).getWikiPageDictionary()
					.getUid(realName);
	}

	public synchronized void rename(String uid, String newName, String langCode)
			throws PageRenamingException {
		if (newName.indexOf("/") != -1) {
			throw new PageRenamingException(
					"You can rename only page/subpage name. / is not allowed.");
		} else if (pageExists(newName, langCode)) {
			throw new PageRenamingException(
					"Page with the name you're trying rename to already exists.");
		}

		if (uid.indexOf("/") != -1) {
			// renaming subpage - create aprop. newName
			newName = uid.substring(0, uid.lastIndexOf("/") + 1) + newName;
		}

		if (!arePagesEdited()) {
			languageDataSources.get(langCode).getWikiPageDictionary().rename(
					uid, newName);

			// add opposite translation
			languageDataSources.get(langCode).getWikiPageDictionary().rename(
					newName, uid);

		} else {
			// some pages are edited now - execute watcher
			languageDataSources.get(langCode).getRenamePageWatcher().rename(
					uid, newName);

			// add opposite translation
			languageDataSources.get(langCode).getRenamePageWatcher().rename(
					newName, uid);

			if (!languageDataSources.get(langCode).getRenamePageWatcher()
					.isRuning()) {
				languageDataSources.get(langCode).getRenamePageWatcher()
						.start();
			}
		}
	}

	public boolean uidInDictionary(String uid, String langCode) {
		return languageDataSources.get(langCode).getWikiPageDictionary()
				.uidInDictionary(uid);
	}

	public boolean realNameInDictionary(String realName, String langCode) {
		return languageDataSources.get(langCode).getWikiPageDictionary()
				.realNameInDictionary(realName);
	}

	public boolean deleteAttachment(String pageName, String attName,
			String langCode) {
		return languageDataSources.get(langCode).getAttachmentDataSource()
				.deleteAttachment(getUid(pageName, langCode), attName);
	}

	public boolean deleteAttachments(String pageName, String langCode) {
		return languageDataSources.get(langCode).getAttachmentDataSource()
				.deleteAttachments(getUid(pageName, langCode));
	}

	public Set<String> getAttachmentsSet(WikiPage page, String langCode) {
		return languageDataSources.get(langCode).getAttachmentDataSource()
				.getAttachmentsSet(getUid(page.getName(), langCode));
	}

	public int getLastAttachmentVersion(String pageName, String attachmentName,
			String langCode) {
		return languageDataSources.get(langCode).getAttachmentDataSource()
				.getLastAttachmentVersion(pageName, attachmentName);
	}

	public WikiAttachment getAttachment(String pageName,
			String attachementName, int version, String langCode) {
		return languageDataSources.get(langCode).getAttachmentDataSource()
				.getAttachment(pageName, attachementName, version);
	}

	public WikiAttachment getAttachment(String pageName,
			String attachementName, String langCode) {
		return languageDataSources.get(langCode).getAttachmentDataSource()
				.getAttachment(pageName, attachementName);
	}

	public void addAttachment(InputStream attFile, String attName,
			WikiPage page, String user, String langCode) {
		languageDataSources.get(langCode).getAttachmentDataSource()
				.addAttachment(attFile, attName,
						getUid(page.getName(), langCode), user);
	}

	public Set<String> getAllPageNames(String langCode) {
		Set<String> set = languageDataSources.get(langCode)
				.getMediaDataSource().getAllPageNames();
		Set<String> translatedSet = new TreeSet<String>();

		for (String page : set) {
			translatedSet.add(getRealName(page, langCode));
		}

		return translatedSet;
	}

	public void preSave(String langCode) throws WikiSaveException {
		languageDataSources.get(langCode).getMediaDataSource().preSave();
	}

	public void savePage(Credentials credentials, WikiPage page,
			String langCode, WikiSession wikiSession) throws WikiSaveException,
			EditingNotAllowedException {

		if (!page.isEditable(credentials)) {
			throw new EditingNotAllowedException(
					"This page is not editable for user: "
							+ credentials.getName());
		}

		preSave(langCode);

		if (page.getLastVersion() != page.getVersion()) {
			page.setVersion(page.getLastVersion());
		}

		WikiType postEditProcess = getWikiType("postedit");

		if (postEditProcess != null) {
			// there are some plugins for postediting a page
			page = postEditProcess.process(page, wikiSession);
		}

		// put processed page to pages map
		languageDataSources.get(langCode).getPages().put(page.getName(), page);

		if (langCode.equals(defaultLangugeCode)) {
			languageDataSources.get(langCode).getMediaDataSource().savePage(
					getUid(page.getName(), langCode), page);
		} else {
			languageDataSources.get(langCode).getMediaDataSource().savePage(
					getUid(page.getName(), langCode), page);
		}

		postSave(langCode);

	}

	public void postSave(String langCode) throws WikiSaveException {
		languageDataSources.get(langCode).getMediaDataSource().postSave();
	}

	public boolean preGet(String langCode) {
		return languageDataSources.get(langCode).getMediaDataSource().preGet();
	}

	public boolean postGet(String langCode) {
		return languageDataSources.get(langCode).getMediaDataSource().postGet();
	}

	public WikiPage getPageAtVersion(WikiPage originPage, boolean loadContent,
			String languageCode, int version) {
		return languageDataSources.get(languageCode).getMediaDataSource()
				.getPageAtVersion(originPage,
						getUid(originPage.getName(), languageCode),
						loadContent, version);
	}

	public WikiPage getPageAtVersion(WikiPage originPage, boolean loadContent,
			int version, String langCode) {
		return languageDataSources.get(langCode).getMediaDataSource()
				.getPageAtVersion(originPage,
						getUid(originPage.getName(), langCode), loadContent,
						version);
	}

	public void getContentAtVersion(WikiPage page, boolean loadContent,
			int version, String langCode) {
		languageDataSources.get(langCode).getMediaDataSource()
				.getContentAtVersion(page, getUid(page.getName(), langCode),
						loadContent, version);
	}

	public int getPageMod(String pageName, String langCode) {
		return languageDataSources.get(langCode).getMediaDataSource()
				.getPageMod(pageName);
	}

	public void setPageMod(String pageName, int mods, String langCode) {
		languageDataSources.get(langCode).getMediaDataSource().setPageMod(
				pageName, mods);
	}

	public boolean deletePage(String pageName, String langCode) {
		languageDataSources.get(langCode).getMediaDataSource().setPageMod(
				getUid(pageName, langCode),
				MediaDataSource.EDITABLE + MediaDataSource.VIEWABLE);

		return languageDataSources.get(langCode).getMediaDataSource()
				.deletePage(getUid(pageName, langCode));
	}

	public long getAttachmentSize(String pageName, String attachmentName,
			int version, String langCode) {
		return languageDataSources.get(langCode).getAttachmentDataSource()
				.getAttachmentSize(pageName, attachmentName, version);
	}

	public synchronized int getEditSessionId() {
		return editSessionId++;
	}

	public Integer getSessionIdExpireTime() {
		return sessionIdExpireTime;
	}

	public boolean arePagesEdited() {
		return (editedPages.size() > 0);
	}

	public void endEditing(Integer editSessionId) throws PageNotEditedException {
		if (!editedPages.contains(editSessionId)) {
			throw new PageNotEditedException("Page with sessionid "
					+ editSessionId + " is not beeing edited.");
		}

		editedPages.remove(editSessionId);
	}

	public void addEditing(Integer editSessionId) {
		editedPages.add(editSessionId);
	}

	/*
	 * public Semaphore getS() { return s; }
	 */

	public HashMap<String, String> getPagesWaiting(String langCode) {
		return languageDataSources.get(langCode).getRenamePageWatcher()
				.getPagesWaiting();
	}

	public Credentials getCredentialsInstance(Object param) {
		Credentials cred = null;

		log.debug("Getting instance of " + credentialsClass.getName());

		try {
			cred = (Credentials) credentialsClass.newInstance();
			cred.setParam(param);
		} catch (InstantiationException e) {
			log.error("Couldn't get instance of class "
					+ credentialsClass.getName(), e);
		} catch (IllegalAccessException e) {
			e.printStackTrace();
		}

		return cred;
	}

	public void saveRolesSet(WikiPage page, Set<String> roles, int action,
			String langCode) throws DataSourceException {
		languageDataSources.get(langCode).getMediaDataSource().saveRolesSet(
				getUid(page.getName(), langCode), roles, action);
	}

	private String unmarshallText(Node textNode) {
		StringBuffer buf = new StringBuffer();

		Node n;
		NodeList nodes = textNode.getChildNodes();

		for (int i = 0; i < nodes.getLength(); i++) {
			n = nodes.item(i);

			if (n.getNodeType() == Node.TEXT_NODE) {
				buf.append(n.getNodeValue());
			}
		}

		return buf.toString();
	}

	public Map<String, String> getAllLanguageCodes() {
		return allLanguageCodes;
	}

	public Map<String, String> getUsedLanguageCodes() {
		return usedLanguageCodes;
	}

	private void loadLanguage(String langCode) {
		MediaDataSource newMDS = null;
		AttachmentDataSource newADS = null;
		WikiPageDictionary newWPD = null;

		Properties jbosswikiprops = new Properties();

		try {
			jbosswikiprops.load(WikiEngine.class.getResourceAsStream("/"
					+ PROPERTIES_FILENAME));
		} catch (IOException e) {
			e.printStackTrace();
		}

		// LOAD data sources
		if (jbosswikiprops.getProperty("mediaDataSourceClass") != null) {
			try {
				newMDS = (MediaDataSource) Class.forName(
						jbosswikiprops.getProperty("mediaDataSourceClass"))
						.newInstance();
			} catch (InstantiationException e) {
				log.error(e);
			} catch (IllegalAccessException e) {
				log.error(e);
			} catch (ClassNotFoundException e) {
				log.error(e);
			}
			log.info("Loaded mediaDataSource: " + newMDS);

		} else {
			// if sth went wrong
			newMDS = new FileMediaDataSource();
		}

		newMDS.setLanguageCode(langCode);

		if (jbosswikiprops.getProperty("attachmentDataSourceClass") != null) {
			try {
				newADS = (AttachmentDataSource) Class
						.forName(
								jbosswikiprops
										.getProperty("attachmentDataSourceClass"))
						.newInstance();
			} catch (InstantiationException e) {
				log.error(e);
			} catch (IllegalAccessException e) {
				log.error(e);
			} catch (ClassNotFoundException e) {
				log.error(e);
			}

			log.info("Loaded attachmentDataSource: " + newADS);
		} else {
			// if sth went wrong
			newADS = new FileAttachmentDataSource();
		}

		newADS.setLanguageCode(langCode);

		if (jbosswikiprops.getProperty("wikiPageDictionaryClass") != null) {
			try {
				newWPD = (WikiPageDictionary) Class.forName(
						jbosswikiprops.getProperty("wikiPageDictionaryClass"))
						.newInstance();
			} catch (InstantiationException e) {
				log.error(e);
			} catch (IllegalAccessException e) {
				log.error(e);
			} catch (ClassNotFoundException e) {
				log.error(e);
			}
			log.info("Loaded wikiPageDictionary: " + newWPD);
		} else {
			// if sth went wrong
			newWPD = new FileWikiPageDictionary();
		}

		newWPD.setLanguageCode(langCode);

		// set wikiEngine in classes
		newMDS.setWikiEngine(this);
		newADS.setWikiEngine(this);
		newWPD.setWikiEngine(this);

		LanguageDataSource newLDS = new LanguageDataSource(this, langCode,
				newMDS, newADS, newWPD);

		languageDataSources.put(langCode, newLDS);

		usedLanguageCodes.put(langCode, allLanguageCodes.get(langCode));
	}

	/**
	 * Adds new language structure to current datasource.
	 * 
	 * @param langCode
	 *            two-letter language code
	 * @param langName
	 *            language description
	 */
	public void addNewLanguage(String langCode, String langName) {

		loadLanguage(langCode);

		defaultDS().getMediaDataSource().addLangCode(langCode, langName);

	}

	public void saveMetaProps(Properties props, WikiPage page) {
		languageDataSources
				.get(page.getLangCode())
				.getMediaDataSource()
				.saveMetadataProps(
						languageDataSources.get(page.getLangCode())
								.getWikiPageDictionary().getUid(page.getName()),
						props);

		page.setMetaDataProps(props);
	}

	public boolean wikiTypeExists(String wikiTypeName) {
		return (wikiTypes.get(wikiTypeName) != null);
	}

	public Properties getMetaDataProps(String realName, String langCode) {
		return languageDataSources.get(langCode).getMediaDataSource()
				.getMetadataProps(getUid(realName, langCode));
	}

	public void removeDictForPage(String pageName, String langCode)
			throws PageRenamingException {
		if (!arePagesEdited()) {
			languageDataSources.get(langCode).getWikiPageDictionary()
					.removeDictForPage(getUid(pageName, langCode));

			// remove opposite translation
			languageDataSources.get(langCode).getWikiPageDictionary()
					.removeDictForPage(pageName);
		} else {
			// there are some pages edited - execute watcher
			languageDataSources.get(langCode).getRenamePageWatcher()
					.removeDictForPage(getUid(pageName, langCode));

			// remove opposite translation
			languageDataSources.get(langCode).getRenamePageWatcher()
					.removeDictForPage(pageName);

			if (!languageDataSources.get(langCode).getRenamePageWatcher()
					.isRuning()) {
				languageDataSources.get(langCode).getRenamePageWatcher()
						.start();
			}
		}
	}

	public Set<String> getDictRemovesWaiting(String langCode) {
		return languageDataSources.get(langCode).getRenamePageWatcher()
				.getDictRemovesWaiting();
	}

	public Map<String, String> getRenameMapings(String langCode) {
		return languageDataSources.get(langCode).getWikiPageDictionary()
				.getRenameMapings();
	}

	/**
	 * Sets global property
	 * 
	 * @param key
	 *            Key
	 * @param value
	 *            Value
	 * @param langCode
	 *            Language code
	 */
	public void setGlobalProperty(String key, String value, String langCode) {
		try {
			languageDataSources.get(langCode).getMediaDataSource()
					.setGlobalProperty(key, value);
		} catch (DataSourceException e) {
			log.error(e);
		}
	}

	/**
	 * Gets global property
	 * 
	 * @param key
	 *            Key
	 * @param langCode
	 *            Language code
	 * @return Value
	 */
	public String getGlobalProperty(String key, String langCode) {
		try {
			return languageDataSources.get(langCode).getMediaDataSource()
					.getGlobalProperty(key);
		} catch (DataSourceException e) {
			log.error(e);
			return null;
		}
	}
}
